"""Collection of file format converters."""

import os
import mne
import shlex
import logging
import mne_bids
import mne_nirs

from ..utils import run_shell


def dcm2nii(files, out, dst, **kwargs):
    """Convert DICOM files to NIfTI and write to disk.

    Parameters
    ----------
    files : List of Files
        List of File objects that represent the files to be converted.

    out : str
        Output format desired.

    dst: str
        Destination directory where converted files will be stored.

    config: str
        Path to config file. Should be compatible with dcm2bids
        version installed.

    Returns
    -------
    None
    """

    # Set and build command
    flags = kwargs.get("flags", None)
    tmp = [
        "dcm2bids",
        "-d",
        "{path}",
        "-p",
        "{subject}",
        "-s",
        "{session}",
        "-c",
        "{config}",
        "-o",
        "{dst}",
        "-l",
        "{verbose}",
    ]
    if flags and not flags.issubset(_DCM2NII_VALID_FLAGS):
        raise ValueError(
            "Unsupported flag found \n"
            f"Valid flags: {','.join(_DCM2NII_VALID_FLAGS)}",
        )
    if flags:
        tmp.extend(flags)
    tmp = shlex.join(tmp)

    # Set logging level
    verbose = kwargs.get("logging", "INFO")

    # Set configuration file
    config = kwargs.get("config", None)
    if not config:
        raise KeyError("Expected config path, but not found")

    # Set argument mappping
    tag_map = kwargs.get("tag_map", None)
    if not tag_map:
        logging.info("Using default tag map")
        tag_map = {"subject": "subject", "session": "session"}

    logging.info(f"Using command template: {tmp}")

    for file in files:
        # Fill command with arguments
        args = {}
        for param, tag in tag_map.items():
            tag = file.tags.get(tag, None)
            if tag:
                args[param] = tag.value
        args["path"] = file.path
        cmd = tmp.format(**args, dst=dst, verbose=verbose, config=config)
        run_shell(cmd)
        logging.info(f"Converted {file.path} and stored to {dst}")
    logging.info("Conversion to nii complete")


def edf2asc(files, out, dst, **kwargs):
    """Convert Eye track data format and write to disk."""

    # Set and build command
    tmp = ["edf2asc", "{path}", "{new_path}"]
    flags = kwargs.get("flags", None)
    if flags and not flags.issubset(_EDF2ASC_VALID_FLAGS):
        raise ValueError(
            "Unsupported flag found. \n"
            f"Valid flags: {', '.join(_EDF2ASC_VALID_FLAGS)}"
        )
    if flags:
        tmp.extend(flags)
    tmp = shlex.join(tmp)

    logging.info(f"Using command template: {tmp}")

    for file in files:
        # Fill command with arguments
        args = {}
        new_path = os.path.join(
            dst, file.build_modified_path(extension="asc", sourcetype=None)
        )
        args["path"] = file.path
        args["new_path"] = new_path
        os.makedirs(os.path.dirname(new_path), exist_ok=True)
        cmd = tmp.format(**args)
        run_shell(cmd)
        logging.info(f"Converted {file.path} and stored to {new_path}")
    logging.info("Conversion to asc complete")


def nirs_conv(files, out, dst, **kwargs):
    """Convert NIRS data format and write to disk."""

    for file in files:
        raw = mne.io.read_raw_nirx(file.path)
        raw.anonymize(**kwargs.get("anonymize", {}))
        new_path = os.path.join(
            dst,
            file.build_modified_path(
                extension="snirf",
                sourcetype=None,
            ),
        )
        os.makedirs(os.path.dirname(new_path), exist_ok=True)
        mne_nirs.io.write_raw_snirf(raw, new_path)
        logging.info(f"Converted {file.path} and stored to {new_path}")
    logging.info("Conversion to snirf complete")


def eeg_conv(files, out, dst, **kwargs):
    """Convert EEG data format and write to disk."""

    # Set conversion logging level
    verbose = kwargs.get("logging", "INFO")

    # Set power line frequency to default if not provided
    line_freq = kwargs.get("line_freq", 50)
    logging.info(f"Using value of {line_freq} Hz for power line frequency")

    # Set BIDSPath map
    tag_map = kwargs.get("tag_map", None)
    if not tag_map:
        logging.info("Using default tag map")
        tag_map = {
            "subject": "subject",
            "session": "session",
            "acquisition": "acquisition",
            "task": "task",
            "datatype": "datatype",
            "run": "run",
            "suffix": "suffix",
        }

    # Set overwrite preferences
    overwrite = kwargs.get("overwrite", True)
    logging.info(f"Using overwrite = {overwrite} for conversion")

    # Set format preferences
    logging.info(f"Convertion will be to {out} format")

    # Set event code mappings
    event_id = kwargs.get("events_map", None)
    if not event_id:
        logging.info("Continuing without events code map")

    # Set anonymization strategy
    anonymize = kwargs.get("anonymize", None)
    if not anonymize:
        logging.info("Continuing without data anonymization")
    else:
        logging.info(f"Using daysback of {anonymize['daysback']} to anonymize")

    for file in files:
        raw = mne.io.read_raw(file.path)
        raw.info["line_freq"] = line_freq

        # Form BIDSPath parameters from tag mapped_column
        path_params = {}
        for param, tag in tag_map.items():
            tag = file.tags.get(tag, None)
            path_params[param] = tag.value if tag else None

        path = mne_bids.BIDSPath(root=dst, **path_params)

        mne_bids.write_raw_bids(
            raw,
            path,
            event_id=event_id,
            anonymize=anonymize,
            format=out,
            overwrite=overwrite,
            verbose=verbose,
        )

        logging.info(f"Converted {file.path} and stored to {path}")
    logging.info(f"Conversion to {out} format complete")


def convert(files, out, dst, **kwargs):
    """Convert a file collection from one format to another."""

    # Proceed only if all files have the same extension
    extension = set()
    for file in files:
        extension.add(file.tags.get("extension").value)
    if len(extension) > 1:
        raise TypeError(
            "Expected all files of same extension, but received different \n"
            f"Found {len(extension)}: {', '.join(extension)}",
        )

    # Check if output format supported and choose converter
    extension = extension.pop()
    for outs in _SUPPORTED:
        if out in outs and extension in _SUPPORTED[outs]["from"]:
            converter = _SUPPORTED[outs]["using"]
            break

    if not converter:
        raise ValueError(f"Extension {extension} or output {out} mismatch.")

    if not dst or dst == "":
        raise ValueError("Expected valid destination path, received empty")

    # Initiate conversion
    logging.info(f"Converting {extension} to {out} using {converter.__name__}")
    converter(files, out, dst, **kwargs)


_SUPPORTED = {
    (
        "BrainVision",
        "EDF",
        "FIF",
        "auto",
    ): {
        "from": (
            ".bdf",
            ".cnt",
            ".data",
            ".edf",
            ".gdf",
            ".mat",
            ".mff",
            ".nxe",
            ".set",
            ".vhdr",
        ),
        "using": eeg_conv,
    },
    ("ASCII",): {"from": (".edf",), "using": edf2asc},
    ("NIfTI",): {"from": (".dcm",), "using": dcm2nii},
    ("SNIRF",): {"from": (".nirx",), "using": nirs_conv},
}
_DCM2NII_VALID_FLAGS = {"--forceDcm2niix", "--clobber"}
_EDF2ASC_VALID_FLAGS = {
    "-t",
    "-c",
    "-z",
    "-v",
    "-y",
    "-sp",
    "-sh",
    "-sg",
    "-l",
    "-nr",
    "-r",
    "-nl",
    "-res",
    "-vel",
    "fvel",
    "-s",
    "-ne",
    "-e",
    "-ns",
    "-nv",
    "-nst",
    "-nmsg",
    "-neye",
    "-nflags",
    "-hpos",
    "-avg",
    "-ftime",
    "-input",
    "-buttons",
    "-failsafe",
    "-ntarget",
    "-ntime_check",
    "-npa_check",
    "-logmsg",
}



from .db import DBManager
from .data import Indexer
from .data import File, Layout, Tag
from .data import Specification

__all__ = [
    "DBManager",
    "File",
    "Layout",
    "Tag",
    "Indexer",
    "Specification",
]



"""Define the standard for the data set."""

import re
import os
import logging
import warnings

import pandas as pd
from string import Formatter

from .. import utils


_TAG_PATTERN = re.compile(r"({([\w\d]*?)(?:<([^>]+)>)?(?:\|((?:\.?[\w])+))?\})")

__all__ = ["Specification"]


class Specification:
    """Representation of the specification used.

    Attributes
    ----------
    name : str
        The name of the specification.

    Parameters
    ----------
    spec : dict or str
        The dict-converted YAML specification file or its path.

    Notes
    -----
    The YAML file that describes the specification can be split into two parts.
       - tag definitions
       - path definitions

    More information on these will follow.
    """

    def __init__(self, details=None, name=None):
        self.spec, self.name = get_spec(details, name)

    def get_valid_tags(self):
        """Returns valid tag names."""
        return [t.get("name") for t in self.spec.get("tags")]

    def extract_tags(self, path):
        """Returns tag:value pairs for file."""
        tags = {}
        for tag in self.spec.get("tags"):
            val = re.findall(tag.get("pattern"), path)
            if val:
                tags[tag.get("name")] = val[0]
        return tags

    def validate(self, path):
        """Returns True if path is valid."""
        tags = self.extract_tags(path)
        if self.build_path(tags) == path:
            return True
        return False

    def build_path(self, tags, strict=False):
        """Construct path provided a set of tags.

        Parameters
        ----------
        tags : dict
            A dictionary with tag name, value pairs.

        strict : bool
            If True, the tags provided should exactly match the
            requirements. If False, extra tags are ignored and the
            first matching path is built.

        Returns
        -------
        path : str
            The constructed path if successful, else `None`.

        """

        path_patterns = self.spec.get("path_patterns")
        logging.debug(f"Building path with tags : {tags}")

        # Remove none values
        tags = {k: v for k, v in tags.items() if v or v == 0}

        # Work with extension with or without .
        if "extension" in tags:
            ext = tags.get("extension")
            tags["extension"] = ext if ext.startswith(".") else "." + ext

        # Attempt to match pattern with tags and return first match
        for pattern in path_patterns:
            path = pattern
            matches = _TAG_PATTERN.findall(pattern)

            # Do not tamper with tags provided so that
            # it can be used for other patterns
            tags_copy = tags.copy()

            # Skip if strict set and all tags not matched
            if strict and set(tags_copy.keys()) - set([t[1] for t in matches]):
                continue

            # Validate and fill in missing tags with default
            for subpattern, name, valid_vals, default in matches:
                valid = [v for v in valid_vals.split("|")]

                if valid and name in tags_copy and tags_copy[name] not in valid:
                    continue

                if name not in tags_copy and default:
                    tags_copy[name] = default

                if valid and default and default not in valid:
                    warnings.warn(
                        f"Pattern {subpattern} is inconsistent as it defines an invalid default value"
                    )

                # Simplify path
                path = path.replace(subpattern, "{%s}" % name)

            # Keep or remove optional tags
            optional_patterns = re.findall(r"(\[.*?\])", path)
            for op in optional_patterns:
                optional_tag = re.findall(r"\{(.*?)\}", op)[0]
                path = (
                    path.replace(op, op[1:-1])
                    if optional_tag in tags_copy.keys()
                    else path.replace(op, "")
                )

            # Find fields available in path
            fields = [f[1] for f in Formatter().parse(path)]

            # Proceed only if all field data available
            if set(fields) - set(tags_copy.keys()):
                continue

            # Fill in the fields
            path = path.format_map(tags_copy)

            return path

        return None

    def organize(self, rules):
        """Create a organized copy of a source directory based on rules.

        Parameters
        ----------
        rules : dict
            A dictionary describing the instructions for organizing.

        Returns
        -------
        None
        """

        # Check for required keys
        for mandatory_key in [
            "source",
            "destination",
            "pattern",
            "tag_rules",
        ]:
            if mandatory_key not in rules:
                raise KeyError(f"Expected '{mandatory_key}' key in rule.")

        source = rules.get("source")
        destination = rules.get("destination")
        logging.info(f"Initiating organization of {source} -> {destination}")

        overwrite = rules.get("overwrite", False)
        if overwrite:
            logging.warning("Overwrite set: Existing files may be overwritten")

        add = rules.get("add", None)
        for a in add or []:
            logging.info(f"File {a['path']} will be added as {a['position']}.")

        logging.debug(f"Matching contents with pattern {rules.get('pattern')}")

        # Organize matched files as per rules
        matches = utils.get_dir_contents(
            source, rules.get("pattern"), rules.get("skip", None)
        )
        for file in matches or []:
            logging.info(f"Found match with file {file}")

            # Extract tags
            tags = {}
            for rule in rules.get("tag_rules"):
                tag_name = rule.get("name")
                logging.debug(f"Foraging for {tag_name} tag")

                if "value" in rule:
                    val = rule.get("value")
                    tags[tag_name] = val
                    logging.debug(f"Setting tag with {val}")
                else:
                    match = re.findall(rule.get("pattern"), file)
                    if match and len(match) != 1:
                        logging.warning("Expected single match, found more.")

                    # Choose last match always
                    val = match[-1] if match else None
                    logging.debug(f"Matching with pattern yields {val}")

                    if "prepend" in rule and val:
                        val = "".join([str(rule.get("prepend")), val])
                        logging.debug(f"Prepending tag value to get {val}")

                    if "length" in rule and val and len(val) != rule.get("length"):
                        if "iffy_prepend" in rule:
                            logging.debug("Insufficient length, prepending")
                            val = "".join([str(rule.get("iffy_prepend")), val])

                        if len(val) != rule.get("length"):
                            logging.debug("Tag value of insufficient length")
                            val = None

                    if c := rule.get("case") in ["lower", "upper"] and val:
                        val = val.lower() if c == "lower" else val.upper()

                    if "default" in rule and not val:
                        val = rule.get("default")
                        logging.debug(f"Using default value of {val} for tag")

                    if "padding" in rule and val:
                        pad = rule.get("padding")

                        # Set defaults
                        direction = (pad.get("direction", "left"),)
                        char = (pad.get("char", "0"),)
                        length = pad["length"]

                        if direction == "left":
                            val.rjust(length, char)

                        elif direction == "right":
                            val.ljust(length, char)

                    if "replace" in rule and val:
                        rep = rule.get("replace")
                        col, with_, from_ = [rep[x] for x in ["col", "with", "from"]]
                        logging.info(f"File {from_} will be used to map tag values")
                        mapping = pd.read_csv(from_, dtype=str)

                        # TODO: Modularize below code snippet
                        m = mapping.where(mapping[col] == val).dropna()

                        if len(m) == 0:
                            logging.error(f"No mapping found for {val}")
                            continue

                        if len(m) > 1:
                            logging.error(f"Expected unique map for {val}, found many")
                            continue

                        val = m[with_].values[0]

                    if not val:
                        logging.error(f"Value for {tag_name} tag not found in {file}.")

                logging.info(f"File marked with {tag_name}:{val} tag")
                tags.update({tag_name: val})

            # Warning, clunky code ahead. To be made better
            rel_path = self.build_path(tags)
            if not rel_path:
                logging.error("Unable to build destination path for file")
                continue

            new_path = os.path.join(rules.get("destination"), rel_path)
            logging.info(f"Target destination path is {new_path}")
            utils.copy(file, new_path, overwrite)
            logging.info("Moved file to target")

            if add:
                for addition in add:
                    if addition["position"] == "content":
                        addition_path = os.path.join(
                            new_path, os.path.basename(addition["path"])
                        )
                    elif addition["position"] == "fellow":
                        addition_path = os.path.join(
                            os.path.dirname(new_path), addition["path"]
                        )
                    else:
                        raise ValueError(
                            "Expected position to be either content or fellow"
                        )
                    utils.copy(addition["path"], addition_path, overwrite)
                    logging.info(f"Added addition at {addition_path}")

            if not rules.get("copy_fellows", False):
                continue

            # Find fellows
            logging.info("Initiating copying of fellow files")
            fellows = [
                f.path
                for f in os.scandir(os.path.dirname(file))
                if f.name != os.path.basename(file) and not f.is_dir()
            ]
            logging.info(f"Found {len(fellows)} fellows accompanying the file")

            for fellow in fellows:
                # Tag changes for fellow
                logging.info(f"Changing tags for fellow {fellow}")
                tags_copy = tags.copy()
                tags_copy.update({"extension": os.path.splitext(fellow)[1]})
                for rule in rules.get("rename_rules", []):
                    if re.findall(rule.get("target"), fellow):
                        tag_val = rule.get("suffix")
                        tags_copy.update({"suffix": tag_val})
                        logging.info(f"File marked with suffix:{tag_val} tag")

                # Copy fellow files
                rel_path = self.build_path(tags_copy)
                if not rel_path:
                    logging.error(f"Unable to build destination path for file {file}")
                    continue
                new_path = os.path.join(rules.get("destination"), rel_path)
                logging.info(f"Target destination path is {new_path}")
                utils.copy(fellow, new_path, overwrite)

    def __repr__(self):
        return f"<Specification name={self.name}>"


def get_spec(path=None, name=None):
    if not path:
        path = os.path.join(os.path.dirname(__file__), "configs", "bids.yaml")

    spec = utils.read_yaml(path)
    name = name if name else os.path.splitext(os.path.basename(path))[0]

    return spec, name



"""Database and DataFrame functionality for data access and manipulation."""

import os
import logging
import pandas as pd

from sqlalchemy import Table
from sqlalchemy import Column
from sqlalchemy import MetaData
from sqlalchemy import ForeignKey
from sqlalchemy import ForeignKeyConstraint

from sqlalchemy.types import Boolean
from sqlalchemy.types import Float
from sqlalchemy.types import Integer
from sqlalchemy.types import String
from sqlalchemy.types import DateTime

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from .utils import get_dtype, log_df, log_col

__all__ = ["DBManager", "common_records", "migrate", "transform", "validate"]


class DBManager:
    """Interface to connect with db and perform operations."""

    def __init__(self, db_path=None):
        self.engine = get_db(db_path)
        self.meta = self._init_metadata()
        self._sessionmaker = sessionmaker(self.engine)
        self._session = None

    def _init_metadata(self):
        meta = MetaData()
        meta.reflect(bind=self.engine)
        return meta

    @property
    def session(self):
        if self._session is None:
            self._session = self._sessionmaker()
        return self._session

    @property
    def connection(self):
        return self.engine.connect()

    def build_column(self, name, dtype, **kwargs):
        """Build SQLalchemy Column object given column description."""
        primary = kwargs.get("primary", False)
        fk = [ForeignKey(f)] if (f := kwargs.get("refs")) else []
        return Column(name, self.get_type(dtype), *fk, primary_key=primary)

    def build_constraint(self, cols, links):
        """Build SQLalchmy constraint objects given links."""
        return ForeignKeyConstraint(cols, links)

    def create_table(self, table, cols, refs=[]):
        """Create or extend table in db given the description."""
        cls = [self.build_column(**c) for c in cols]
        cns = [self.build_constraint(**r) for r in refs]
        table = Table(table, self.meta, *cls, *cns, extend_existing=True)
        table.create(bind=self.engine, checkfirst=True)

    def get_primary(self, table):
        """Returns priamry keys for table."""
        return [c.name for c in self.meta.tables[table].primary_key]

    def get_type(self, dtype, default_length=250):
        """
        Return supported SQLalchemy type equivalent of provided dtype string.

        Parameters
        ----------
        dtype: str
            Supported dtype string representation.

        default_length: int, optional
            Default length of string. Used if required by db backend.
        """

        SQL_TYPE_EQUIVALENT = {
            "boolean": Boolean,
            "datetime": DateTime,
            "float": Float,
            "integer": Integer,
            "str": String,
        }

        dtype, length = get_dtype(dtype, default_length)
        stype = SQL_TYPE_EQUIVALENT[dtype]

        return stype(length) if length else stype()

    def get_table(self, table, cols=None):
        """
        Retrieve table records in db as a DataFrame.

        Parameters
        ----------
        table : str
            Table in database from which to retrieve records.

        cols : list of str
            Column names to select from table.
        """
        return pd.read_sql_table(table, self.connection, columns=cols)

    def to_table(
        self,
        df,
        table,
        check_dups=True,
        resolve_dups=False,
        check_fks=True,
        resolve_fks=False,
        insert_ignore=False,
        drop_na=None,
        threshold=None,
        if_exists="append",
        index=False,
        insert_method=None,
        **kwargs,
    ):
        """Write records in DataFrame to a table.

        Parameters
        ----------
        df : pandas.DataFrame
            DataFrame containing records.

        table : str
            Table in database into which the records will be inserted.

        check_dups : bool, default True
            Check for duplicates in records.

        resolve_dups : [False, 'first', 'last'], default False
            Resolution method for duplicates if found.

        check_fks : bool, default False
            Check if foreign keys present in parent.

        resolve_fks : bool, default False
            Attempt to resolve missing foreign keys by inserting to parent.

        insert_ignore : bool, default Flase
            Ignore insertion of records already present in table.

        drop_na : list of df column names, default None
            If provided, records with na in all given columns are dropped.

        threshold : int, None
            If non-NA values < threshold, then record dropped.

        if_exists : ['append', 'replace', 'fail'], default 'append'
            Insert behavior in case table exists.

            - 'append' : Insert new values to the existing table.
            - 'replace' : Drop the table before inserting new values.
            - 'fail' : Raise a ValueError if table exists.

        index : bool, default False
            Write DataFrame index as a column. Uses index_label as the
            column name in the table.

        insert_method : {None, 'multi', callable}, optional
            Controls the SQL insertion clause used.

            - None : Uses standard SQL INSERT clause (one per row).
            - ‘multi’: Pass multiple values in a single INSERT clause.
            - callable with signature ``(pd_table, conn, keys, data_iter)``.

            Details and a sample callable implementation can be found
            in the pandas section `insert method
            <https://pandas.pydata.org/docs/user_guide/io.html#io-sql-method>`_.

        kwargs : key, value mappings Other keyword arguments are
            passed down to `pandas.DataFrame.to_sql()
            <https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_sql.html>`_

        Returns
        -------
        rows : None or int
            Number of rows affected by to_sql. None is returned if the
            callable passed into `insert_method` does not return an
            integer number of rows.

        Raises
        ------
        ValueError
            - When values provided are not sufficient for insert operation.
            - When the table already exists and `if_exists` is 'fail'.

        OperationalError
            - Most likely there are duplicates records in the
              DataFrame. Other reasons are related to the database
              operation and are detailed in sqlalchemy section
              `OperationalError
              <https://docs.sqlalchemy.org/en/20/errors.html#operationalerror>`_.

        """

        if drop_na or threshold:
            logging.info("Dropping records with missing information")
            df = df.dropna(subset=drop_na, thresh=threshold)

        if check_dups:
            df = df[self.resolve_dups(df, table, resolve_dups)]

        if check_fks:
            df = df[self.resolve_fks(df, table, resolve_fks)]

        if insert_ignore:
            mask = common_records(df, self.get_table(table).astype(df.dtypes))
            logging.info(f"Ignoring insert of {mask.sum()} common records")
            df = df[~mask]

        logging.info(f"Inserting {len(df.index)} records")

        df.to_sql(
            table,
            self.connection,
            if_exists=if_exists,
            index=index,
            method=insert_method,
            **kwargs,
        )

    def resolve_dups(self, df, table, resolve=False):
        """
        Resolve duplicate primary keys in DataFrame.

        Parameters
        ----------
        df : pandas.DataFrame
            DataFrame to check for duplicate records.
        table : str
            Table the records will be inserted into.
        resolve : [False, 'first', 'last'], default False
            Determines resolution method.

            * False : Mark all duplicates as False.
            * 'first' : Mark duplicates as False except for first occurence.
            * 'last' : Mark duplicates as False except for last occurence.

        Returns
        -------
        mask : pandas.Series
            Series of booleans showing whether each record in the
            Dataframe is not a duplicate.

        Raises
        ------
        ValueError
            When primary key duplicates are found and `resolve` is True.

        """
        dups = df.duplicated(self.get_primary(table), resolve)
        log_df(df[dups], "Found duplicate records")
        return ~dups

    def resolve_fks(self, df, table, resolve=False):
        """
        Parameters
        ----------
        df : pandas.DataFrame
            Non-duplicated records to be checked.
        table : str
            Table the records will be inserted into.
        resolve : bool, optional
            If True, attempt to resolve missing parent records by inserting.

        Returns
        -------
        mask : pandas.Series
             Series of booleans showing whether each record in the
             DataFrame has all required parent records.

        Warnings
        --------
        Resolution fails with a ValueError when a foreign key
        constraint on a table does not reference all the columns that
        provide the required values to insert records into the parent
        table.

        """
        mask = pd.Series(True, index=df.index)

        for fkc in self.meta.tables[table].foreign_key_constraints:
            p_cols = [fk.column.name for fk in fkc.elements]
            p_df = self.get_table(fkc.referred_table.name, p_cols)
            p_mask = common_records(df, p_df, fkc.column_keys, p_cols)

            log_df(df[~p_mask], "Missing parent records")

            if not p_mask.all() and resolve:
                logging.info("Resolving missing records by insert to parent")
                self.to_table(
                    df[~p_mask][fkc.column_keys].set_axis(p_cols, axis=1),
                    fkc.referred_table.name,
                    check_dups=True,
                    resolve_dups="first",
                    check_fks=True,
                    resolve_fks=True,
                )
                p_mask[~p_mask] = True

            mask &= p_mask
        return mask

    def __repr__(self):
        return f"<DBManager '{self.engine.url}'>"


def get_db(db_path):
    """Returns SQLalchemy engine for provided URL."""
    if not db_path:
        db_path = "sqlite:///{}".format(
            os.path.join(os.path.expanduser("~"), "index.sqlite")
        )
    return create_engine(db_path)


def common_records(child, parent, child_on=None, parent_on=None):
    """Check whether each record in a DataFrame is contained in another."""

    return (
        pd.merge(
            child.reset_index(),
            parent,
            how="left",
            left_on=child_on,
            right_on=parent_on,
            indicator="exists",
        )
        .replace({"both": True, "left_only": False, "right_only": False})
        .set_index("index")["exists"]
        .astype("bool")
    )


def check_for_key(key, mapping):
    """Return columns in table mapping which contain the key provided."""

    return [c["name"] for c in mapping["cols"] if key in c.keys()]


def migrate(
    src,
    target,
    mapping,
    dry_run=False,
    na_vals=[],
    dtype_kws=None,
    **kwargs,
):
    """
    Transforms and migrates records from one db to another.

    Parameters
    ----------
    src: str
        SQLAlchemy database URL of database from which data is
        transferred.

    target: str
        SQLAlchemy database URL of database to which data is
        transferred.

    mapping: dict
        Dictionary providing information on column mappings, type,
        transformations, and validation checks.

    dry_run: bool
        If dry run, no records are inserted into target and tables are
        not created, but errors and invalid records are logged.

    na_vals: list, optional
        List of values to be considered as record not available. By
        default, the values '', 'None', 'NONE', 'NA', and 'Not Applicable'
        are considered.

    dtype_kws : dict, optional
        Key, value pairs that will be passed to
        :meth:`napi.db.set_dtype()`.

    kwargs : key, value mappings
        Other keyword arguments are passed down to
        :meth:`napi.db.to_table`.
    """

    s = DBManager(src)
    t = DBManager(target)

    na = ["", "None", "NONE", "NA", "N/A", "<NA>", "Not applicable"] + na_vals

    for m in mapping:
        logging.info(f"Transferring table {m['maps']} -> {m['table']}")

        # Extract source records
        records = s.get_table(m["maps"]).replace(na, pd.NA)

        # Transform and validate
        records = transform(records, dtype_kws, m)
        mask = validate(records, m)

        # Reshape data records
        df = reshape(records[mask], m.get("reshape", dict()))

        if dry_run:
            continue

        # Load records into target
        t.create_table(m["table"], m["cols"] + m.get("attach", []), m.get("refs", []))
        t.to_table(df, m["table"], threshold=m.get("threshold"), **kwargs)


def set_dtype(series, dtype, **kwargs):
    """
    Set dtype of series.

    Parameters
    ----------
    series: pandas.Series
        Series for which the dtype has to set.

    dtype: str
        Supported dtype to which the series will be converted.

    kwargs: key, value mappings
        Other keyword arguments are passed down to `pandas.to_datetime()`_
        if the dtype is 'datetime'.
    """

    dtype, _ = get_dtype(dtype)

    if dtype == str:
        series = series.astype(dtype)

    elif dtype == "datetime":
        series = pd.to_datetime(series, errors="coerce", **kwargs)

    elif dtype in {"float", "integer"}:
        series = pd.to_numeric(series, errors="coerce", downcast=dtype)

    elif dtype == "boolean":
        series = series.replace(
            {
                "1": True,
                "0": False,
                "True": True,
                "False": False,
                "Yes": True,
                "No": False,
            }
        )
        series = series.astype(dtype)

    return series.convert_dtypes()


def reshape(records, steps):
    """Reshape table records into appropriate shape."""

    for procedure in steps:
        [(p, k)] = procedure.items()

        if p == "add":
            records[k["name"]] = k["value"]

        if p == "split":
            records[k["rename"]] = records[k["name"]].str.split(k["pat"], expand=True)

        if p == "melt":
            records = records.melt(**k)

        if p == "pivot":
            records = records.pivot_table(**k).reset_index()

    return records


def transform(records, dtype_kws, mapping):
    """Transform table records into appropriate format."""

    df = pd.DataFrame(index=records.index)
    to_hide = check_for_key("hide", mapping)
    error = pd.Series(False, index=records.index)

    for col in mapping["cols"] + mapping.get("detach", []):
        name, maps = col["name"], col["maps"]
        logging.debug(f"Transforming column {name}")
        df[name], col_error = transform_column(records[maps], dtype_kws, **col)
        error |= col_error

    log_df(records[error], "Error transforming records", hide=to_hide)
    return df[~error]


def replace_value(value, field, using, file):
    # Get mapping
    mapping = pd.read_csv(file, dtype=str)

    # Find value to replace
    result = mapping.query("`%s` == @value" % field)

    if len(result) == 0:
        logging.error(f"Value '{value}' not found in field '{field}' of file '{file}'.")
        return

    if len(result) > 1:
        logging.error(f"Non-unique mappings for value '{value}' in file {file}")
        return

    return result.at[0, using]


def replace_column(series, field, using, file, strict=True):
    # Load mapping from file
    mapping = pd.read_csv(file, dtype=str)
    logging.info(f"Replacing values in '{field}' with '{using}' from {file}")

    # Stop if non-unique mappings
    if mapping.duplicated([field]).any():
        raise ValueError(f"Non-unique mappings found in file {file}")

    mapping = pd.Series(mapping[using], index=mapping[field])
    replaced = series.map(mapping)

    # Retain original value if replacement not strict
    if not strict:
        replaced.fillna(series, inplace=True)

    return replaced


def transform_column(series, dtype_kws=dict(), **kwargs):
    """Transform column to appropriate datatype."""

    hide = kwargs.get("hide", False)
    s, error = series.copy(deep=True), series.isna()

    if pat := kwargs.get("extract"):
        s = s.astype(str).str.extract(pat, expand=False)
        logging.info(f"Extracting and replacing based on pattern {pat}")

    if rep := kwargs.get("replace"):
        if "field" in rep:
            replace_column(s, **rep)
            logging.info("Replacing values with mapping in external file")
        else:
            s = s.replace({k["value"]: k["with"] for k in rep})
            logging.info("Replacing values with given in config")

    if ca := kwargs.get("case"):
        s = s.str.upper() if ca == "upper" else s.str.lower()
        logging.info(f"Changing case to {ca}")

    s = set_dtype(s, kwargs["dtype"], **dtype_kws)
    error = series.notna() & s.isna()
    log_col(series[error], f"Error transforming values to {kwargs['dtype']}", hide=hide)

    return s, error


def validate(records, mapping):
    """Validate table records and return mask where True means valid."""

    to_hide = check_for_key("hide", mapping)
    mask = pd.Series(True, index=records.index)

    for col in mapping["cols"] + mapping.get("detach", []):
        logging.debug(f"Validating column {col['name']}")
        mask &= validate_column(records[col["name"]], **col)

    log_df(records[~mask], "Found invalid records", hide=to_hide)
    return mask


def validate_column(series, **kwargs):
    """Validate column and return mask where True means valid."""

    hide = kwargs.get("hide", False)
    mask = pd.Series(True, index=series.index)

    if kwargs.get("primary"):
        mask &= (m := series.notna())
        log_col(series[~m], "Primary column values cannot be NA", hide=hide)

    if pat := kwargs.get("like"):
        mask &= (m := series.astype(str).str.fullmatch(pat) | series.isna())
        log_col(series[~m], f"Values do not match pattern {pat}", hide=hide)

    if bounds := kwargs.get("between"):
        mask &= (m := series.between(*bounds) | series.isna())
        log_col(series[~m], f"Values not between bounds {bounds}", hide=hide)

    if members := kwargs.get("in"):
        mask &= (m := series.isin(members) | series.isna())
        log_col(series[~m], f"Values not in {members}", hide=hide)

    return mask
